from meiga import Error, Result, isSuccess

from lume.src.domain.services.killer_service import KillerService


class FakeKillerService(KillerService):
    def execute(self, process: str) -> Result[bool, Error]:
        print("executing")
        return isSuccess



import os
import time
from typing import List, Tuple

import requests
from meiga import Error, Failure, Result, Success, isFailure, isSuccess
from meiga.decorators import meiga

from lume.config import Config
from lume.src.application.use_cases.env_manager import EnvManager
from lume.src.application.use_cases.messages import get_colored_command_message
from lume.src.domain.services.executor_service import ExecutorService
from lume.src.domain.services.killer_service import KillerService
from lume.src.domain.services.logger import (
    COMMAND,
    ERROR,
    HIGHLIGHT,
    INFO,
    WAITING,
    WARNING,
    Logger,
)
from lume.src.domain.services.setup_service import SetupService


class EmptyConfigError(Error):
    pass


class ActionNotFoundError(Error):
    pass


class CwdIsNotADirectoryError(Error):
    pass


def on_empty_config(self, action):
    self.logger.log(WARNING, f"Empty config for {action}")


def on_error_with_cwd(self, action):
    self.logger.log(
        WARNING, f"Given execution directory (cwd) for {action} action is not valid"
    )


SETUP_DETACH_DEFAULT_LOG_FILENAME = "lume.setup_detach.log"


class LumeUseCase:
    def __init__(
        self,
        config: Config,
        executor_service: ExecutorService,
        detach_executor_service: ExecutorService,
        detach_killer_service: KillerService,
        setup_service: SetupService,
        logger: Logger,
    ):
        self.config = config
        self.executor_service = executor_service
        self.detach_executor_service = detach_executor_service
        self.detach_killer_service = detach_killer_service
        self.setup_service = setup_service
        self.logger = logger
        self.env_manager = EnvManager(self.logger)
        self.steps = None

    @meiga
    def execute(self, steps: List[str]):
        self.env_manager.set(self.config.shared_envs)
        self.config.check_requirements().unwrap_or_return()
        self.steps = steps
        for step in self.steps:
            self.logger.log(HIGHLIGHT, f"Step: {step}")
            if step == "setup":
                result = self.setup_service.execute()
                if result.is_failure:
                    self.logger.log(ERROR, f"Setup: {result.value.message}")
                result.unwrap_or_return()
            else:
                cwd = (
                    self._get_cwd(step)
                    .handle(on_failure=on_error_with_cwd, failure_args=(self, step))
                    .unwrap_or_return()
                )

                self._set_env(step)
                processes = (
                    self._run_setup_detach(step, cwd)
                    .handle(on_failure=self._run_teardown, failure_args=(cwd, step))
                    .unwrap_or([])
                )
                self._run_setup(step, cwd).handle(
                    on_failure=self._run_teardown_detach, failure_args=processes
                ).handle(
                    on_failure=self._run_teardown, failure_args=(cwd, step)
                ).unwrap_or_return()
                self._run_commands(step, cwd, processes).unwrap_or_return()
                self._run_teardown_detach(processes)
                self._run_teardown(cwd, step)
                self._unset_env(step)

        self.env_manager.unset(self.config.shared_envs)

        return isSuccess

    def clear_env(self):
        self.env_manager.unset(self.config.shared_envs)
        if self.steps:
            for step in self.steps:
                if step != "setup":
                    self._unset_env(step)

    def _set_env(self, action):
        if action == "install":
            step = self.config.install
        elif action == "uninstall":
            step = self.config.uninstall
        else:
            step = self.config.steps.get(action)
        self.env_manager.set_step(step)

    def _unset_env(self, action):
        if action == "install":
            step = self.config.install
        elif action == "uninstall":
            step = self.config.uninstall
        else:
            step = self.config.steps.get(action)
        self.env_manager.unset_step(step)

    @meiga
    def _run_setup(self, step, cwd) -> Result:
        setup_commands = self._get_setup_commands(step).unwrap_or([])
        for setup_command in setup_commands:
            message = get_colored_command_message(
                setup_command, cwd, step, prefix="setup"
            )
            self.logger.log(COMMAND, message)
            self.executor_service.execute(setup_command, cwd).unwrap_or_return()
        return isSuccess

    @meiga
    def _run_setup_detach(self, step, cwd) -> Result[List, Error]:
        setup_detach_commands, log_filename = self._get_setup_detach_commands(
            step
        ).unwrap_or(([], SETUP_DETACH_DEFAULT_LOG_FILENAME))
        processes = []
        for setup_detach_command in setup_detach_commands:
            message = get_colored_command_message(
                setup_detach_command, cwd, step, prefix="setup-detach"
            )
            self.logger.log(COMMAND, message)
            process = self.detach_executor_service.execute(
                setup_detach_command, cwd, log_filename
            ).unwrap()
            if process:
                processes.append(process)
        return Success(processes)

    def _run_teardown_detach(self, processes):
        if processes:
            for process in processes:
                self.detach_killer_service.execute(process)

    def _wait_if_necessary(self, action):
        if action in ["install", "uninstall"]:
            return
        else:
            step = self.config.steps.get(action)
            if not step.wait_seconds and not step.wait_http_200:
                return
            if step.wait_seconds:
                self.logger.log(WAITING, f"Waiting {step.wait_seconds} seconds")
                time.sleep(step.wait_seconds)

            if step.wait_http_200:
                self.logger.log(WAITING, f"Waiting for 200 -> {step.wait_http_200}")
                wait_seconds_retry = float(
                    os.environ.get("LUME_WAIT_HTTP_200_WAIT_SECONDS_RETRY", 1)
                )
                num_max_attempts = int(
                    os.environ.get("LUME_WAIT_HTTP_200_NUM_MAX_ATTEMPTS", 20)
                )

                is_ok = False
                num_attempts = 0
                for i in range(num_max_attempts):
                    num_attempts += 1
                    try:
                        response = requests.get(step.wait_http_200)
                        status = response.status_code
                        status_message = f"{status}        \033[F"
                        if status == 200:
                            is_ok = True
                            break
                        else:
                            status_message = f"Unexpected return code -> {status} | {response.text} \033[F"
                    except:  # noqa E722
                        status_message = "Connection Error\033[F"

                    self.logger.log(
                        INFO, f"  Attempt {num_attempts} -> {status_message}"
                    )
                    time.sleep(wait_seconds_retry)

                time_elapsed = round((wait_seconds_retry * num_attempts), 2)
                if is_ok:
                    self.logger.log(
                        INFO,
                        f"  Received a 200 after {num_attempts} attempts in ~{time_elapsed} seconds",
                    )
                else:
                    self.logger.log(
                        WARNING,
                        f"  Not received any 200 after {num_attempts} attempts in ~{time_elapsed} seconds",
                    )

        return

    @meiga
    def _run_commands(self, step, cwd, processes) -> Result:
        self._wait_if_necessary(step)
        commands = (
            self._get_commands(step)
            .handle(on_failure=on_empty_config, failure_args=(self, step))
            .unwrap_or([])
        )
        for command in commands:
            message = get_colored_command_message(command, cwd, step)
            self.logger.log(COMMAND, message)
            self.executor_service.execute(command, cwd).handle(
                on_failure=self._run_teardown_detach, failure_args=processes
            ).handle(
                on_failure=self._run_teardown, failure_args=(cwd, step)
            ).unwrap_or_return()

        return isSuccess

    def _run_teardown(self, cwd, step):
        teardown_commands = self._get_teardown_commands(step).unwrap_or([])
        for teardown_command in teardown_commands:
            message = get_colored_command_message(
                teardown_command, cwd, step, prefix="teardown"
            )
            self.logger.log(COMMAND, message)
            self.executor_service.execute(teardown_command, cwd)

    def _get_commands(self, action) -> Result[List[str], Error]:
        if action == "install":
            if not self.config.install:
                return Failure(EmptyConfigError())
            commands = self.config.install.run
        elif action == "uninstall":
            if not self.config.uninstall:
                return Failure(EmptyConfigError())
            commands = self.config.uninstall.run
        else:
            step = self.config.steps.get(action)
            if not step:
                return Failure(EmptyConfigError())
            commands = step.run

        return Success(commands)

    def _get_setup_commands(self, action) -> Result[List[str], Error]:
        if action in ["install", "uninstall"]:
            return isFailure
        else:
            step = self.config.steps.get(action)
            if not step:
                return Failure(EmptyConfigError())
            setup_commands = step.setup
            if not setup_commands:
                return isFailure

        return Success(setup_commands)

    def _get_setup_detach_commands(
        self, action
    ) -> Result[Tuple[List[str], str], Error]:
        if action in ["install", "uninstall"]:
            return isFailure
        else:
            step = self.config.steps.get(action)
            if not step:
                return Failure(EmptyConfigError())
            setup_detach = step.setup_detach
            if not setup_detach or not setup_detach.get("run"):
                return isFailure

            return Success(
                (
                    setup_detach.get("run"),
                    setup_detach.get("log_filename", f"lume.{action}.setup_detach.log"),
                )
            )

    def _get_teardown_commands(self, action) -> Result[List[str], Error]:
        if action in ["install", "uninstall"]:
            return isFailure
        else:
            step = self.config.steps.get(action)
            if not step:
                return Failure(EmptyConfigError())
            teardown_commands = step.teardown
            if not teardown_commands:
                return isFailure

        return Success(teardown_commands)

    def _get_cwd(self, action) -> Result[str, Error]:
        if action == "install":
            if not self.config.install:
                return Success(None)
            cwd = self.config.install.cwd
            if cwd and not os.path.isdir(cwd):
                return Failure(CwdIsNotADirectoryError())
            return Success(cwd)
        if action == "uninstall":
            if not self.config.uninstall:
                return Success(None)
            cwd = self.config.uninstall.cwd
            if cwd and not os.path.isdir(cwd):
                return Failure(CwdIsNotADirectoryError())
            return Success(cwd)
        else:
            if not self.config.steps:
                return Success(None)
            step = self.config.steps.get(action)
            if step:
                cwd = step.cwd
                if cwd and not os.path.isdir(cwd):
                    return Failure(CwdIsNotADirectoryError())
                return Success(cwd)
            else:
                return Failure(ActionNotFoundError())



from lume.src.domain.services.logger import (
    COMMAND,
    ENVAR,
    ENVAR_WARNING,
    ERROR,
    GLOBAL,
    HIGHLIGHT,
    INFO,
    WAITING,
    WARNING,
    Logger,
)
from lume.src.infrastructure.services.logger.colors import Colors

LOGGING_LEVEL = {
    HIGHLIGHT: "🟩",
    COMMAND: "💻",
    GLOBAL: "🌐",
    INFO: "",
    ERROR: "💩",
    WARNING: "⚠️",
    ENVAR: "  🔸",
    ENVAR_WARNING: "  🔸",
    WAITING: "⏱ ",
}


class EmojisLogger(Logger):
    def log(self, logging_level, message):
        start, end = self.color_provider(logging_level)
        print(f"{start}{LOGGING_LEVEL[logging_level]} {message.rstrip()}{end}")

    @staticmethod
    def color_provider(logging_level):
        start = ""
        end = ""
        if logging_level == ERROR:
            start = Colors.FAIL
        elif logging_level == WARNING or logging_level == ENVAR_WARNING:
            start = Colors.WARNING
        elif logging_level == HIGHLIGHT:
            start = Colors.OKBLUE

        if start != "":
            end = Colors.ENDC

        return start, end



from typing import Optional

from pydantic import BaseModel


class DependencyConfig(BaseModel):
    type: str
    url: str
    unzip: bool
    auth_required: Optional[bool] = False
    credentials_env: Optional[str] = None
    overwrite: Optional[bool] = False



import argparse
import os
import platform
import shutil
import sys
import time
import traceback

import yaml
from meiga import Error, Failure, Result, Success, isFailure, isSuccess

from lume import __version__
from lume.config import Config
from lume.config.config_file_not_found_error import ConfigFileNotFoundError
from lume.config.config_file_not_valid_error import ConfigFileNotValidError
from lume.config.required_env_error import RequiredEnvError
from lume.src.application.use_cases.use_case_builder import UseCaseBuilder
from lume.src.infrastructure.services.logger.colors import Colors


def has_args(args):
    is_active = False
    for arg in vars(args):
        if arg == "no_strict":
            continue
        is_active = is_active or getattr(args, arg)
    return is_active


def get_config(filename: str = r"lume.yml") -> Result[Config, Error]:
    if not os.path.isfile(filename):
        return Failure(ConfigFileNotFoundError(filename))

    try:
        with open(filename) as file:
            lume_dict = yaml.load(file, Loader=yaml.FullLoader)
            config = Config(lume_dict)
            return Success(config)
    except Exception as e:  # noqa
        message = f"Error loading {filename} file: {repr(e.__class__)} {e} | {traceback.format_exc()}"
        return Failure(ConfigFileNotValidError(message))


def on_config_failure(result: Result, config_file: str):
    print(f"❌  Cannot load lume configuration from: {config_file}")
    print(
        "❌  If you aren't using lume in the same directory as a lume.yml file, please use LUME_CONFIG_FILENAME env var to configure it"
    )
    if isinstance(result.value, ConfigFileNotValidError):
        print(f"❌  {Colors.FAIL}{result.value}{Colors.ENDC}")


def on_execution_failure(result: Result):
    if isinstance(result.value, RequiredEnvError):
        unmeet_required_env_messages = result.value.unmeet_required_env_messages
        for env, description in unmeet_required_env_messages.items():
            print(
                f"❌  {Colors.FAIL}{env}{Colors.ENDC} environment variable is mandatory ➜ {description}"
            )
        print(
            f"❌  Please, review required env variables defined in {Colors.OKGREEN}lume.yml{Colors.ENDC} ({Colors.OKBLUE}required_env{Colors.ENDC})"
        )


def get_parser(config):
    parser = argparse.ArgumentParser(
        prog="lume 🔥",
        description="Lume helps you with your daily dev operations and ease the CI & CD process.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "-v", "--version", action="store_true", help="show lume version number."
    )
    parser.add_argument(
        "-all",
        "--all-commands",
        action="store_true",
        dest="all_commands",
        help="run all commands",
    )
    for command in config.get_commands():
        parser.add_argument(
            f"-{command}",
            f"--{command}",
            action="store_true",
            dest=f"command_{command}",
            help=f"{command}",
        )
    parser.add_argument(
        "--no-strict",
        action="store_true",
        dest="no_strict",
        help="No fails if not exist",
    )
    return parser


def str2bool(value: str) -> bool:
    return value.lower() in [
        "true",
        "1",
        "t",
        "y",
        "yes",
        "yeah",
        "yup",
        "certainly",
        "uh-huh",
    ]


def get_strict_mode(args):
    no_strict_mode = str2bool(os.getenv("LUME_NO_STRICT", "false"))
    if args.no_strict or no_strict_mode:
        return False
    else:
        return True


def check_command_availability(strict_mode, not_known, parser, config_file) -> Result:
    if len(not_known) > 0:
        not_supported_message = f"lume 🔥: Given commands are not supported ({not_known}). Please, check your lume file ({config_file})"

        if not strict_mode:
            print(not_supported_message)
            print(
                "lume 🌈: As you define the '--no-strict' (or use `LUME_NO_STRICT` envvar) option everything is ok and return code is 0"
            )
            return isSuccess
        else:
            parser.print_help()
            print(f"\n{not_supported_message}")
            return isFailure


def main():
    start = time.time()
    header = f" 🔥 lume {__version__} ({platform.system()} -- Python {platform.python_version()}) "
    columns = shutil.get_terminal_size().columns
    print(header.center(columns - 10, "="))
    result = isFailure
    exit_code = 1
    suffix = "(exit code 1)"
    prefix = f"{Colors.FAIL} Failed"

    config_file = os.environ.get("LUME_CONFIG_FILENAME", "lume.yml")
    config = get_config(filename=config_file).unwrap_or_else(
        on_failure=on_config_failure, failure_args=(Result.__id__, config_file)
    )

    if config:
        parser = get_parser(config)
        args, not_known = parser.parse_known_args()
        strict_mode = get_strict_mode(args)
        config.update_strict_mode(strict_mode)
        lume_use_case = UseCaseBuilder.lume(config=config)

        if not has_args(args):
            result = check_command_availability(
                strict_mode, not_known, parser, config_file
            )
        else:
            dict_args = vars(args)

            if args.version:
                print(f"lume 🔥 => {__version__}")
                return

            selected_actions = [
                action
                for action, selected in dict_args.items()
                if selected and action != "all_commands"
                if selected and action != "no_strict"
            ]

            if args.all_commands:
                all_steps_actions = [
                    action
                    for action in dict_args.keys()
                    if "command_" in action
                    and action not in ["command_install", "command_uninstall"]
                ]
                selected_actions += all_steps_actions

            selected_actions = [
                action.replace("command_", "") for action in selected_actions
            ]
            result = lume_use_case.execute(steps=selected_actions).handle(
                on_failure=on_execution_failure, failure_args=(Result.__id__)
            )
            lume_use_case.clear_env()

        if result.is_success:
            exit_code = 0
            suffix = "(exit code 0)"
            prefix = f"{Colors.OKGREEN} Succeed"

        if not config.settings["show_exit_code"]:
            suffix = ""
    else:
        suffix = ""

    end = time.time()
    elapsed_time = end - start

    footer = f"{prefix} in {elapsed_time:.2f} seconds {suffix} {Colors.ENDC}".center(
        len(header), "="
    )
    print(footer.center(columns, "="))

    sys.exit(exit_code)



from abc import ABC, abstractmethod

from meiga import Error, NotImplementedMethodError, Result


class SetupService(ABC):
    @abstractmethod
    def execute(self) -> Result[bool, Error]:
        return NotImplementedMethodError



from lume.config import Config
from lume.src.application.use_cases.lume_use_case import LumeUseCase
from lume.src.infrastructure.services.executor.detach_popen_executor_service import (
    DetachPopenExecutorService,
)
from lume.src.infrastructure.services.executor.popen_executor_service import (
    PopenExecutorService,
)
from lume.src.infrastructure.services.killer.popen_killer_service import (
    PopenKillerService,
)
from lume.src.infrastructure.services.logger.emojis_logger import EmojisLogger
from lume.src.infrastructure.services.setup.setup_service import BucketSetupService


class UseCaseBuilder:
    @staticmethod
    def lume(config: Config):
        default_logger = EmojisLogger()
        default_executor_service = PopenExecutorService(logger=default_logger)
        default_detach_executor_service = DetachPopenExecutorService(
            logger=default_logger
        )
        default_detach_killer_service = PopenKillerService(logger=default_logger)
        setup_config = config.steps["setup"] if "setup" in config.steps else None
        default_setup_service = BucketSetupService(
            setup_config=setup_config, logger=default_logger
        )

        return LumeUseCase(
            config=config,
            executor_service=default_executor_service,
            detach_executor_service=default_detach_executor_service,
            detach_killer_service=default_detach_killer_service,
            setup_service=default_setup_service,
            logger=default_logger,
        )



from typing import Dict

from meiga import Error


class RequiredEnvError(Error):
    def __init__(self, unmeet_required_env_messages: Dict[str, str]):
        self.unmeet_required_env_messages = unmeet_required_env_messages



from typing import Dict, List, Optional

from pydantic import BaseModel

from lume.config.check_list_or_str_item import check_list_or_str_item
from lume.config.check_os_list_or_str_item import check_os_list_or_str_item
from lume.config.get_envs import get_envs


class StepConfig(BaseModel):
    run: List[str]
    cwd: Optional[str] = None
    envs: Dict[str, str] = dict()
    setup: Optional[List[str]] = None
    teardown: Optional[List[str]] = None
    setup_detach: Optional[Dict] = None
    wait_seconds: Optional[int] = None
    wait_http_200: Optional[str] = None
    overwrote_envs: List[str] = list()

    def add_shared_env(self, shared_envs: Dict[str, str]):
        if shared_envs and self.envs:
            self.overwrote_envs = list(
                set(shared_envs.keys()).intersection(set(self.envs))
            )

    @staticmethod
    def from_dict(kdict):
        run = check_os_list_or_str_item(kdict, "run", required=True)
        setup = check_os_list_or_str_item(kdict, "setup", required=False)
        teardown = check_os_list_or_str_item(kdict, "teardown", required=False)
        setup_detach = kdict.get("setup_detach")
        if setup_detach:
            setup_detach["run"] = check_list_or_str_item(
                setup_detach, "run", required=True, suffix=" (setup_detach)"
            )
        envs = get_envs(kdict)
        return StepConfig(
            run=run,
            cwd=kdict.get("cwd"),
            envs=envs,
            setup=setup,
            teardown=teardown,
            setup_detach=setup_detach,
            wait_seconds=kdict.get("wait_seconds"),
            wait_http_200=kdict.get("wait_http_200"),
        )
